
using System;
using System.IO;
using System.Collections;

namespace at.jku.ssw.Analyse {

public class Token {
	public int kind;    // token kind
	public int pos;     // token position in the source text (starting at 0)
	public int col;     // token column (starting at 0)
	public int line;    // token line (starting at 1)
	public string val;  // token value
	public Token next;  // ML 2005-03-11 Tokens are kept in linked list
	                    // Count of characters (whitespace, comments, code) and
	                    // count of lines (whitespace, comments, code) from after the end of
	                    // the last token up to the end of this one.
	public LinesOfCodeInfo loc;
}

public class Buffer {
	public const char EOF = (char)256;
	const int MAX_BUFFER_LENGTH = 64 * 1024; // 64KB
	byte[] buf;         // input buffer
	int bufStart;       // position of first byte in buffer relative to input stream
	int bufLen;         // length of buffer
	int fileLen;        // length of input stream
	int pos;            // current position in buffer
	Stream stream;      // input stream (seekable)
	bool isUserStream;  // was the stream opened by the user?
	
	public Buffer (Stream s, bool isUserStream) {
		stream = s; this.isUserStream = isUserStream;
		fileLen = bufLen = (int) s.Length;
		if (stream.CanSeek && bufLen > MAX_BUFFER_LENGTH) bufLen = MAX_BUFFER_LENGTH;
		buf = new byte[bufLen];
		bufStart = Int32.MaxValue; // nothing in the buffer so far
		Pos = 0; // setup buffer to position 0 (start)
		if (bufLen == fileLen) Close();
	}
	
	~Buffer() { Close(); }
	
	public void Close() {
		if (!isUserStream && stream != null) {
			stream.Close();
			stream = null;
		}
	}
	
	public int Read () {
		if (pos < bufLen) {
			return buf[pos++];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos++];
		} else {
			return EOF;
		}
	}

	public int Peek () {
		if (pos < bufLen) {
			return buf[pos];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos];
		} else {
			return EOF;
		}
	}
	
	public string GetString (int beg, int end) {
		int len = end - beg;
		char[] buf = new char[len];
		int oldPos = Pos;
		Pos = beg;
		for (int i = 0; i < len; i++) buf[i] = (char) Read();
		Pos = oldPos;
		return new String(buf);
	}

	public int Pos {
		get { return pos + bufStart; }
		set {
			if (value < 0) value = 0; 
			else if (value > fileLen) value = fileLen;
			if (value >= bufStart && value < bufStart + bufLen) { // already in buffer
				pos = value - bufStart;
			} else if (stream != null) { // must be swapped in
				stream.Seek(value, SeekOrigin.Begin);
				bufLen = stream.Read(buf, 0, buf.Length);
				bufStart = value; pos = 0;
			} else {
				pos = fileLen - bufStart; // make Pos return fileLen
			}
		}
	}
}

public class Scanner {
	const char EOL = '\n';
	const int eofSym = 0; /* pdt */
	const int charSetSize = 256;
	const int maxT = 129;
	const int noSym = 129;
	short[] start = {
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,143, 35,153,  0,150,144, 18, 59, 60,149,147, 53,176,140,148,
	139,137,137,137,137,137,137,137,137,137,141, 54,152,142,151, 62,
	138,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
	  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 57,  0, 58,146,  1,
	  0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
	  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 55,145, 56, 61,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  -1};


	public Buffer buffer; // scanner buffer
	
	private CodeInfo ciCur; // der vom Parser übergebenen aktuelle CodeInfo-Bereich
	private LinesOfCodeInfo locTmp; // die LinesOfCodeInfo, in die temporär die loc-Infos zum aktuellen Token
	                                // gespeichert werden
	private int iLineTmp, iChrTmp;  // Dient zum Ausgleichen von Fehlern mit dem Look-Ahead-Character des Scanners
	private int iOldTokenLine = 0; // zum Zählen der Codezeilen verwendete Variable
	private int ciPos;		// zum Erkennen von Änderungen der Position wegen Kommentaren
	private bool bPosOk = true; // War die aktuelle Position ok, oder schon doppelt gelesen?
	private bool bIgnoreEOL = false; // Soll ein EOF ignoriert werden?
									// wird benötigt, um EOF's hinter Kommentaren richtig zu behandeln
	private bool bEOFOccured = false; // wird benötigt, um mehrfache Aufrufe von NextCh() am Dateiende zu erkennen
	private Parser parser;	// die Instanz des Parsers, um auf im Parser definierte Konstanten zugreifen zu können
	Token lastTok = null;	// Das letzte zurückgegeben Token, um Look-Ahead-Token des Parsers auszugleichen

	
	Token t;          // current token
	char ch;          // current input character
	int pos;          // column number of current character
	int line;         // line number of current character
	int lineStart;    // start position of current line
	int oldEols;      // EOLs that appeared in a comment;
	BitArray ignore;  // set of characters to be ignored by the scanner

	Token tokens;     // list of tokens already peeked (first token is a dummy)
	Token pt;         // current peek token
	
	char[] tval = new char[128]; // text of current token
	int tlen;         // length of current token
	
	public Scanner (CodeInfo ciCur, string fileName) {
		this.ciCur = ciCur;
		try {
			Stream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read);
			buffer = new Buffer(stream, false);
			Init();
		} catch (IOException) {
			Console.WriteLine("--- Cannot open file {0}", fileName);
			System.Environment.Exit(1);
		}
	}
	
	public Scanner (CodeInfo ciCur, Stream s) {
		this.ciCur = ciCur;
		buffer = new Buffer(s, true);
		Init();
	}
	
	// Diese Methode wird vom Parser aufgerufen, wenn der dich der aktuell
	// betrachtet Bereich ändert (z.B. Namespace->Klasse->Methode)
	public void setCurCodeInfo(CodeInfo ciCur) {
		this.ciCur = ciCur;
	}
	
	void Init() {
		pos = -1; line = 1; lineStart = 0;
		
		// Arbeits-LinesOfCodeInfo erstellen
		locTmp = new LinesOfCodeInfo();
		iLineTmp = 0;
		iChrTmp = 0;
		// Am Anfang der Datei befinden wir uns immer in der ersten Zeile
		locTmp.iLines = 1;
		// initialisierung
		ciPos = pos;
		
		oldEols = 0;
		NextCh();
		ignore = new BitArray(charSetSize+1);
		ignore[' '] = true;  // blanks are always white space
		ignore[9] = true; ignore[10] = true; ignore[13] = true; 
		pt = tokens = new Token();  // first token is a dummy
	}
	
	void NextCh() {
		// jetzt können die Infos, die im vorherigen NextCh() gesammelt wurden,
		// zum aktuellen Bereich addiert werden
		locTmp.iLines += iLineTmp;
		locTmp.iChars += iChrTmp;
		iLineTmp = 0;
		iChrTmp = 0;
				
		if (oldEols > 0) {
			// oldEol wird während der Behandlung von Kommentaren erhöht. 
			// War das Kommentar /* */, wird folgendes Flag gesetzt, um
			// Kommentarzeilen nicht fälschlicherweise als Leerzeilen zu zählen
			// (in NextToken())
			if (ch == '/') // bei Kommentaren mit // ist ch == 10 (EOL)
				bIgnoreEOL = true;
			ch = EOL; oldEols--;
		} 
		else {
			ch = (char)buffer.Read(); pos++;

			// wurde der Zähler pos vielleicht geändert?
			// Dies kann auftreten, wenn ein / Zeichen gelesen wurde,
			// auf Kommentare überprüft wurde, aber keine festgestellt wurden
			if (pos > ciPos && !bEOFOccured) { 
				ciPos = pos;
				bPosOk = true;
			}
			else bPosOk = false;
	
			// replace isolated '\r' by '\n' in order to make
			// eol handling uniform across Windows, Unix and Mac
			if (ch == '\r' && buffer.Peek() != '\n') ch = EOL;
			if (ch == EOL) {
				bIgnoreEOL = false; // Normale EOL wieder beachten
				line++; lineStart = pos + 1;
			}
			
			// es wird nur gezählt, wenn der Zähler pos nicht durch
			// Kommentar-Handler verändert wurde und nicht EOF
			// aufgetreten ist. Das verhindert Doppelzählungen
			if (bPosOk) {
				if (ch == EOL) {				
					// die Gesamtzahl der gezählten Zeilen erhöhen
					iLineTmp++;
				}
				else if (ch != '\r' && ch != Buffer.EOF) {
					// die Gesamtzahl der gezählten Zeichen erhöhen
					iChrTmp++;
				}
			}
			// Nach dem Dateiende wird aufgehört, bei nächsten Aufrufen wird bPosOk false
			if (ch == Buffer.EOF) bEOFOccured = true;
		}

	}

	void AddCh() {
		if (tlen >= tval.Length) {
			char[] newBuf = new char[2 * tval.Length];
			Array.Copy(tval, 0, newBuf, 0, tval.Length);
			tval = newBuf;
		}
		tval[tlen++] = ch;
		NextCh();
	}



	bool Comment0() {
		int level = 1, line0 = line, lineStart0 = lineStart;
		NextCh();
		if (ch == '/') {
			NextCh();
			for(;;) {
				if (ch == 10) {
					level--;
					if (level == 0) { oldEols = line - line0; NextCh(); return true; }
					NextCh();
				} else if (ch == Buffer.EOF) return false;
				else NextCh();
			}
		} else {
			if (ch==EOL) {line--; lineStart = lineStart0;}
			pos = pos - 2; buffer.Pos = pos+1; NextCh();
		}
		return false;
	}

	bool Comment1() {
		int level = 1, line0 = line, lineStart0 = lineStart;
		NextCh();
		if (ch == '*') {
			NextCh();
			for(;;) {
				if (ch == '*') {
					NextCh();
					if (ch == '/') {
						level--;
						if (level == 0) { oldEols = line - line0; NextCh(); return true; }
						NextCh();
					}
				} else if (ch == Buffer.EOF) return false;
				else NextCh();
			}
		} else {
			if (ch==EOL) {line--; lineStart = lineStart0;}
			pos = pos - 2; buffer.Pos = pos+1; NextCh();
		}
		return false;
	}


	void CheckLiteral() {
	
		switch (t.val) {
			case "abstract": t.kind = 6; break;
			case "as": t.kind = 7; break;
			case "base": t.kind = 8; break;
			case "bool": t.kind = 9; break;
			case "break": t.kind = 10; break;
			case "byte": t.kind = 11; break;
			case "case": t.kind = 12; break;
			case "catch": t.kind = 13; break;
			case "char": t.kind = 14; break;
			case "checked": t.kind = 15; break;
			case "class": t.kind = 16; break;
			case "const": t.kind = 17; break;
			case "continue": t.kind = 18; break;
			case "decimal": t.kind = 19; break;
			case "default": t.kind = 20; break;
			case "delegate": t.kind = 21; break;
			case "do": t.kind = 22; break;
			case "double": t.kind = 23; break;
			case "else": t.kind = 24; break;
			case "enum": t.kind = 25; break;
			case "event": t.kind = 26; break;
			case "explicit": t.kind = 27; break;
			case "extern": t.kind = 28; break;
			case "false": t.kind = 29; break;
			case "finally": t.kind = 30; break;
			case "fixed": t.kind = 31; break;
			case "float": t.kind = 32; break;
			case "for": t.kind = 33; break;
			case "foreach": t.kind = 34; break;
			case "goto": t.kind = 35; break;
			case "if": t.kind = 36; break;
			case "implicit": t.kind = 37; break;
			case "in": t.kind = 38; break;
			case "int": t.kind = 39; break;
			case "interface": t.kind = 40; break;
			case "internal": t.kind = 41; break;
			case "is": t.kind = 42; break;
			case "lock": t.kind = 43; break;
			case "long": t.kind = 44; break;
			case "namespace": t.kind = 45; break;
			case "new": t.kind = 46; break;
			case "null": t.kind = 47; break;
			case "object": t.kind = 48; break;
			case "operator": t.kind = 49; break;
			case "out": t.kind = 50; break;
			case "override": t.kind = 51; break;
			case "params": t.kind = 52; break;
			case "private": t.kind = 53; break;
			case "protected": t.kind = 54; break;
			case "public": t.kind = 55; break;
			case "readonly": t.kind = 56; break;
			case "ref": t.kind = 57; break;
			case "return": t.kind = 58; break;
			case "sbyte": t.kind = 59; break;
			case "sealed": t.kind = 60; break;
			case "short": t.kind = 61; break;
			case "sizeof": t.kind = 62; break;
			case "stackalloc": t.kind = 63; break;
			case "static": t.kind = 64; break;
			case "string": t.kind = 65; break;
			case "struct": t.kind = 66; break;
			case "switch": t.kind = 67; break;
			case "this": t.kind = 68; break;
			case "throw": t.kind = 69; break;
			case "true": t.kind = 70; break;
			case "try": t.kind = 71; break;
			case "typeof": t.kind = 72; break;
			case "uint": t.kind = 73; break;
			case "ulong": t.kind = 74; break;
			case "unchecked": t.kind = 75; break;
			case "unsafe": t.kind = 76; break;
			case "ushort": t.kind = 77; break;
			case "using": t.kind = 78; break;
			case "virtual": t.kind = 79; break;
			case "void": t.kind = 80; break;
			case "volatile": t.kind = 81; break;
			case "while": t.kind = 82; break;
			case "partial": t.kind = 126; break;
			case "yield": t.kind = 127; break;
			default: break;
		}

		// Diese Methode wird aufgerufen, falls der Scanner eine Zeichenkette erkennt hat
		// und überprüft auf Schlüsselwörter. Da die Methode NextToken() nach dem
		// Aufruf von CheckLiteral() den Token schon zurückgibt, muss die Länge der 
		// Literale hier gezählt werden.
		locTmp.iChCode += tlen;
		// Genauso müssen hier auch die Codezeilen gezählt werden, falls sie mit Literalen beginnen
		if (iOldTokenLine < t.line && t.kind != eofSym && t.kind != noSym) locTmp.iLoCode++;
		iOldTokenLine = t.line;
	}

	Token NextToken() {
		bool wsHelper = false;
		while (ignore[ch]) {
			//hier werden die Whitespace-Zeichen gezählt
			if (bPosOk) {
				if (ch != EOL && ch != '\r') { locTmp.iChWhitespace++; }
				else if (ch == EOL && !bIgnoreEOL) {
					// wenn auf einen EOL nur Whitespaces und dann wieder
					// ein EOL folgt, dann haben wir eine Whitespace-Linie
					if (wsHelper) locTmp.iLoWhitespace++;
					else wsHelper = true;
				}
			}
			NextCh();
		}
		// Damit wird eine Leerzeile direkt vor Dateiende richtig erkannt
		if (ch == Buffer.EOF && wsHelper) locTmp.iLoWhitespace++;
		
		if (ch == '/' && Comment0() ||ch == '/' && Comment1()) return NextToken();
		int apx = 0;
		t = new Token();
		t.pos = pos; t.col = pos - lineStart + 1; t.line = line; 
		int state = start[ch];
		tlen = 0; AddCh();
		
		switch (state) {
			case -1: { t.kind = eofSym; break; } // NextCh already done
			case 0: { t.kind = noSym; break; }   // NextCh already done
			case 1:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'Z' || ch == '_' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 1;}
				else {t.kind = 1; t.val = new String(tval, 0, tlen); CheckLiteral(); return t;}
			case 2:
				if ((ch >= '0' && ch <= '9')) {apx = 0; AddCh(); goto case 2;}
				else if (ch == 'U') {apx = 0; AddCh(); goto case 154;}
				else if (ch == 'u') {apx = 0; AddCh(); goto case 155;}
				else if (ch == 'L') {apx = 0; AddCh(); goto case 156;}
				else if (ch == 'l') {apx = 0; AddCh(); goto case 157;}
				else {
					tlen -= apx;
					pos = pos - apx - 1; line = t.line;
					buffer.Pos = pos + 1; NextCh();
					t.kind = 2; break;}
			case 3:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 4;}
				else {t.kind = noSym; break;}
			case 4:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 4;}
				else if (ch == 'U') {AddCh(); goto case 158;}
				else if (ch == 'u') {AddCh(); goto case 159;}
				else if (ch == 'L') {AddCh(); goto case 160;}
				else if (ch == 'l') {AddCh(); goto case 161;}
				else {t.kind = 2; break;}
			case 5:
				{t.kind = 2; break;}
			case 6:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 6;}
				else if ((ch == 'D' || ch == 'F' || ch == 'M' || ch == 'd' || ch == 'f' || ch == 'm')) {AddCh(); goto case 17;}
				else if ((ch == 'E' || ch == 'e')) {AddCh(); goto case 7;}
				else {t.kind = 3; break;}
			case 7:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 9;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 8;}
				else {t.kind = noSym; break;}
			case 8:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 9;}
				else {t.kind = noSym; break;}
			case 9:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 9;}
				else if ((ch == 'D' || ch == 'F' || ch == 'M' || ch == 'd' || ch == 'f' || ch == 'm')) {AddCh(); goto case 17;}
				else {t.kind = 3; break;}
			case 10:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 10;}
				else if ((ch == 'D' || ch == 'F' || ch == 'M' || ch == 'd' || ch == 'f' || ch == 'm')) {AddCh(); goto case 17;}
				else if ((ch == 'E' || ch == 'e')) {AddCh(); goto case 11;}
				else {t.kind = 3; break;}
			case 11:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 13;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 12;}
				else {t.kind = noSym; break;}
			case 12:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 13;}
				else {t.kind = noSym; break;}
			case 13:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 13;}
				else if ((ch == 'D' || ch == 'F' || ch == 'M' || ch == 'd' || ch == 'f' || ch == 'm')) {AddCh(); goto case 17;}
				else {t.kind = 3; break;}
			case 14:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 16;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 15;}
				else {t.kind = noSym; break;}
			case 15:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 16;}
				else {t.kind = noSym; break;}
			case 16:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 16;}
				else if ((ch == 'D' || ch == 'F' || ch == 'M' || ch == 'd' || ch == 'f' || ch == 'm')) {AddCh(); goto case 17;}
				else {t.kind = 3; break;}
			case 17:
				{t.kind = 3; break;}
			case 18:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '&' || ch >= '(' && ch <= '[' || ch >= ']' && ch <= 255)) {AddCh(); goto case 19;}
				else if (ch == 92) {AddCh(); goto case 162;}
				else {t.kind = noSym; break;}
			case 19:
				if (ch == 39) {AddCh(); goto case 34;}
				else {t.kind = noSym; break;}
			case 20:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 21;}
				else {t.kind = noSym; break;}
			case 21:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 163;}
				else if (ch == 39) {AddCh(); goto case 34;}
				else {t.kind = noSym; break;}
			case 22:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 23;}
				else {t.kind = noSym; break;}
			case 23:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 24;}
				else {t.kind = noSym; break;}
			case 24:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 25;}
				else {t.kind = noSym; break;}
			case 25:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 19;}
				else {t.kind = noSym; break;}
			case 26:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 27;}
				else {t.kind = noSym; break;}
			case 27:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 28;}
				else {t.kind = noSym; break;}
			case 28:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 29;}
				else {t.kind = noSym; break;}
			case 29:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 30;}
				else {t.kind = noSym; break;}
			case 30:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 31;}
				else {t.kind = noSym; break;}
			case 31:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 32;}
				else {t.kind = noSym; break;}
			case 32:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 33;}
				else {t.kind = noSym; break;}
			case 33:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 19;}
				else {t.kind = noSym; break;}
			case 34:
				{t.kind = 4; break;}
			case 35:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '!' || ch >= '#' && ch <= '[' || ch >= ']' && ch <= 255)) {AddCh(); goto case 35;}
				else if (ch == '"') {AddCh(); goto case 51;}
				else if (ch == 92) {AddCh(); goto case 165;}
				else {t.kind = noSym; break;}
			case 36:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 37;}
				else {t.kind = noSym; break;}
			case 37:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '!' || ch >= '#' && ch <= '/' || ch >= ':' && ch <= '@' || ch >= 'G' && ch <= '[' || ch >= ']' && ch <= '`' || ch >= 'g' && ch <= 255)) {AddCh(); goto case 35;}
				else if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 166;}
				else if (ch == '"') {AddCh(); goto case 51;}
				else if (ch == 92) {AddCh(); goto case 165;}
				else {t.kind = noSym; break;}
			case 38:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 39;}
				else {t.kind = noSym; break;}
			case 39:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 40;}
				else {t.kind = noSym; break;}
			case 40:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 41;}
				else {t.kind = noSym; break;}
			case 41:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 35;}
				else {t.kind = noSym; break;}
			case 42:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 43;}
				else {t.kind = noSym; break;}
			case 43:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 44;}
				else {t.kind = noSym; break;}
			case 44:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 45;}
				else {t.kind = noSym; break;}
			case 45:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 46;}
				else {t.kind = noSym; break;}
			case 46:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 47;}
				else {t.kind = noSym; break;}
			case 47:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 48;}
				else {t.kind = noSym; break;}
			case 48:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 49;}
				else {t.kind = noSym; break;}
			case 49:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 35;}
				else {t.kind = noSym; break;}
			case 50:
				if (!(ch == '"') && ch != Buffer.EOF) {AddCh(); goto case 50;}
				else if (ch == '"') {AddCh(); goto case 168;}
				else {t.kind = noSym; break;}
			case 51:
				{t.kind = 5; break;}
			case 52:
				{t.kind = 85; break;}
			case 53:
				{t.kind = 86; break;}
			case 54:
				{t.kind = 87; break;}
			case 55:
				{t.kind = 88; break;}
			case 56:
				{t.kind = 89; break;}
			case 57:
				{t.kind = 90; break;}
			case 58:
				{t.kind = 91; break;}
			case 59:
				{t.kind = 92; break;}
			case 60:
				{t.kind = 93; break;}
			case 61:
				{t.kind = 94; break;}
			case 62:
				{t.kind = 95; break;}
			case 63:
				{t.kind = 97; break;}
			case 64:
				{t.kind = 98; break;}
			case 65:
				{t.kind = 101; break;}
			case 66:
				{t.kind = 102; break;}
			case 67:
				{t.kind = 104; break;}
			case 68:
				{t.kind = 105; break;}
			case 69:
				{t.kind = 107; break;}
			case 70:
				{t.kind = 109; break;}
			case 71:
				{t.kind = 110; break;}
			case 72:
				{t.kind = 112; break;}
			case 73:
				{t.kind = 113; break;}
			case 74:
				{t.kind = 115; break;}
			case 75:
				{t.kind = 117; break;}
			case 76:
				{t.kind = 119; break;}
			case 77:
				{t.kind = 121; break;}
			case 78:
				{t.kind = 123; break;}
			case 79:
				{t.kind = 125; break;}
			case 80:
				if (ch == 'e') {AddCh(); goto case 81;}
				else {t.kind = noSym; break;}
			case 81:
				if (ch == 'f') {AddCh(); goto case 82;}
				else {t.kind = noSym; break;}
			case 82:
				if (ch == 'i') {AddCh(); goto case 83;}
				else {t.kind = noSym; break;}
			case 83:
				if (ch == 'n') {AddCh(); goto case 84;}
				else {t.kind = noSym; break;}
			case 84:
				if (ch == 'e') {AddCh(); goto case 85;}
				else {t.kind = noSym; break;}
			case 85:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 86;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 85;}
				else {t.kind = noSym; break;}
			case 86:
				{t.kind = 130; break;}
			case 87:
				if (ch == 'n') {AddCh(); goto case 88;}
				else {t.kind = noSym; break;}
			case 88:
				if (ch == 'd') {AddCh(); goto case 89;}
				else {t.kind = noSym; break;}
			case 89:
				if (ch == 'e') {AddCh(); goto case 90;}
				else {t.kind = noSym; break;}
			case 90:
				if (ch == 'f') {AddCh(); goto case 91;}
				else {t.kind = noSym; break;}
			case 91:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 92;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 91;}
				else {t.kind = noSym; break;}
			case 92:
				{t.kind = 131; break;}
			case 93:
				if (ch == 'f') {AddCh(); goto case 94;}
				else {t.kind = noSym; break;}
			case 94:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 95;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 94;}
				else {t.kind = noSym; break;}
			case 95:
				{t.kind = 132; break;}
			case 96:
				if (ch == 'f') {AddCh(); goto case 97;}
				else {t.kind = noSym; break;}
			case 97:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 98;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 97;}
				else {t.kind = noSym; break;}
			case 98:
				{t.kind = 133; break;}
			case 99:
				if (ch == 'e') {AddCh(); goto case 100;}
				else {t.kind = noSym; break;}
			case 100:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 101;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 100;}
				else {t.kind = noSym; break;}
			case 101:
				{t.kind = 134; break;}
			case 102:
				if (ch == 'f') {AddCh(); goto case 103;}
				else {t.kind = noSym; break;}
			case 103:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 104;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 103;}
				else {t.kind = noSym; break;}
			case 104:
				{t.kind = 135; break;}
			case 105:
				if (ch == 'i') {AddCh(); goto case 106;}
				else {t.kind = noSym; break;}
			case 106:
				if (ch == 'n') {AddCh(); goto case 107;}
				else {t.kind = noSym; break;}
			case 107:
				if (ch == 'e') {AddCh(); goto case 108;}
				else {t.kind = noSym; break;}
			case 108:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 109;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 108;}
				else {t.kind = noSym; break;}
			case 109:
				{t.kind = 136; break;}
			case 110:
				if (ch == 'r') {AddCh(); goto case 111;}
				else {t.kind = noSym; break;}
			case 111:
				if (ch == 'o') {AddCh(); goto case 112;}
				else {t.kind = noSym; break;}
			case 112:
				if (ch == 'r') {AddCh(); goto case 113;}
				else {t.kind = noSym; break;}
			case 113:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 114;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 113;}
				else {t.kind = noSym; break;}
			case 114:
				{t.kind = 137; break;}
			case 115:
				if (ch == 'a') {AddCh(); goto case 116;}
				else {t.kind = noSym; break;}
			case 116:
				if (ch == 'r') {AddCh(); goto case 117;}
				else {t.kind = noSym; break;}
			case 117:
				if (ch == 'n') {AddCh(); goto case 118;}
				else {t.kind = noSym; break;}
			case 118:
				if (ch == 'i') {AddCh(); goto case 119;}
				else {t.kind = noSym; break;}
			case 119:
				if (ch == 'n') {AddCh(); goto case 120;}
				else {t.kind = noSym; break;}
			case 120:
				if (ch == 'g') {AddCh(); goto case 121;}
				else {t.kind = noSym; break;}
			case 121:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 122;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 121;}
				else {t.kind = noSym; break;}
			case 122:
				{t.kind = 138; break;}
			case 123:
				if (ch == 'e') {AddCh(); goto case 124;}
				else {t.kind = noSym; break;}
			case 124:
				if (ch == 'g') {AddCh(); goto case 125;}
				else {t.kind = noSym; break;}
			case 125:
				if (ch == 'i') {AddCh(); goto case 126;}
				else {t.kind = noSym; break;}
			case 126:
				if (ch == 'o') {AddCh(); goto case 127;}
				else {t.kind = noSym; break;}
			case 127:
				if (ch == 'n') {AddCh(); goto case 128;}
				else {t.kind = noSym; break;}
			case 128:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 129;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 128;}
				else {t.kind = noSym; break;}
			case 129:
				{t.kind = 139; break;}
			case 130:
				if (ch == 'e') {AddCh(); goto case 131;}
				else {t.kind = noSym; break;}
			case 131:
				if (ch == 'g') {AddCh(); goto case 132;}
				else {t.kind = noSym; break;}
			case 132:
				if (ch == 'i') {AddCh(); goto case 133;}
				else {t.kind = noSym; break;}
			case 133:
				if (ch == 'o') {AddCh(); goto case 134;}
				else {t.kind = noSym; break;}
			case 134:
				if (ch == 'n') {AddCh(); goto case 135;}
				else {t.kind = noSym; break;}
			case 135:
				if ((ch == 10 || ch == 13)) {AddCh(); goto case 136;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= 255)) {AddCh(); goto case 135;}
				else {t.kind = noSym; break;}
			case 136:
				{t.kind = 140; break;}
			case 137:
				if ((ch >= '0' && ch <= '9')) {apx = 0; AddCh(); goto case 137;}
				else if (ch == 'U') {apx = 0; AddCh(); goto case 154;}
				else if (ch == 'u') {apx = 0; AddCh(); goto case 155;}
				else if (ch == 'L') {apx = 0; AddCh(); goto case 156;}
				else if (ch == 'l') {apx = 0; AddCh(); goto case 157;}
				else if (ch == '.') {apx++; AddCh(); goto case 169;}
				else if ((ch == 'E' || ch == 'e')) {apx = 0; AddCh(); goto case 14;}
				else if ((ch == 'D' || ch == 'F' || ch == 'M' || ch == 'd' || ch == 'f' || ch == 'm')) {apx = 0; AddCh(); goto case 17;}
				else {t.kind = 2; break;}
			case 138:
				if ((ch >= 'A' && ch <= 'Z' || ch == '_' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 1;}
				else if (ch == '"') {AddCh(); goto case 50;}
				else {t.kind = noSym; break;}
			case 139:
				if ((ch >= '0' && ch <= '9')) {apx = 0; AddCh(); goto case 137;}
				else if (ch == 'U') {apx = 0; AddCh(); goto case 154;}
				else if (ch == 'u') {apx = 0; AddCh(); goto case 155;}
				else if (ch == 'L') {apx = 0; AddCh(); goto case 156;}
				else if (ch == 'l') {apx = 0; AddCh(); goto case 157;}
				else if (ch == '.') {apx++; AddCh(); goto case 169;}
				else if ((ch == 'X' || ch == 'x')) {apx = 0; AddCh(); goto case 3;}
				else if ((ch == 'E' || ch == 'e')) {apx = 0; AddCh(); goto case 14;}
				else if ((ch == 'D' || ch == 'F' || ch == 'M' || ch == 'd' || ch == 'f' || ch == 'm')) {apx = 0; AddCh(); goto case 17;}
				else {t.kind = 2; break;}
			case 140:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 6;}
				else {t.kind = 83; break;}
			case 141:
				if (ch == ':') {AddCh(); goto case 52;}
				else {t.kind = 84; break;}
			case 142:
				if (ch == '=') {AddCh(); goto case 63;}
				else {t.kind = 96; break;}
			case 143:
				if (ch == '=') {AddCh(); goto case 64;}
				else {t.kind = 99; break;}
			case 144:
				if (ch == '=') {AddCh(); goto case 65;}
				else if (ch == '&') {AddCh(); goto case 66;}
				else {t.kind = 100; break;}
			case 145:
				if (ch == '=') {AddCh(); goto case 67;}
				else if (ch == '|') {AddCh(); goto case 68;}
				else {t.kind = 103; break;}
			case 146:
				if (ch == '=') {AddCh(); goto case 69;}
				else {t.kind = 106; break;}
			case 147:
				if (ch == '=') {AddCh(); goto case 70;}
				else if (ch == '+') {AddCh(); goto case 71;}
				else {t.kind = 108; break;}
			case 148:
				if (ch == '=') {AddCh(); goto case 74;}
				else {t.kind = 114; break;}
			case 149:
				if (ch == '=') {AddCh(); goto case 75;}
				else {t.kind = 116; break;}
			case 150:
				if (ch == '=') {AddCh(); goto case 76;}
				else {t.kind = 118; break;}
			case 151:
				if (ch == '=') {AddCh(); goto case 77;}
				else {t.kind = 120; break;}
			case 152:
				if (ch == '=') {AddCh(); goto case 78;}
				else if (ch == '<') {AddCh(); goto case 170;}
				else {t.kind = 122; break;}
			case 153:
				if ((ch == 9 || ch >= 11 && ch <= 12 || ch == ' ')) {AddCh(); goto case 153;}
				else if (ch == 'd') {AddCh(); goto case 80;}
				else if (ch == 'u') {AddCh(); goto case 87;}
				else if (ch == 'i') {AddCh(); goto case 93;}
				else if (ch == 'e') {AddCh(); goto case 171;}
				else if (ch == 'l') {AddCh(); goto case 105;}
				else if (ch == 'w') {AddCh(); goto case 115;}
				else if (ch == 'r') {AddCh(); goto case 123;}
				else {t.kind = noSym; break;}
			case 154:
				if ((ch == 'L' || ch == 'l')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 155:
				if ((ch == 'L' || ch == 'l')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 156:
				if ((ch == 'U' || ch == 'u')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 157:
				if ((ch == 'U' || ch == 'u')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 158:
				if ((ch == 'L' || ch == 'l')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 159:
				if ((ch == 'L' || ch == 'l')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 160:
				if ((ch == 'U' || ch == 'u')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 161:
				if ((ch == 'U' || ch == 'u')) {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 162:
				if ((ch == '"' || ch == 39 || ch == '0' || ch == 92 || ch >= 'a' && ch <= 'b' || ch == 'f' || ch == 'n' || ch == 'r' || ch == 't' || ch == 'v')) {AddCh(); goto case 19;}
				else if (ch == 'x') {AddCh(); goto case 20;}
				else if (ch == 'u') {AddCh(); goto case 22;}
				else if (ch == 'U') {AddCh(); goto case 26;}
				else {t.kind = noSym; break;}
			case 163:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 164;}
				else if (ch == 39) {AddCh(); goto case 34;}
				else {t.kind = noSym; break;}
			case 164:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 19;}
				else if (ch == 39) {AddCh(); goto case 34;}
				else {t.kind = noSym; break;}
			case 165:
				if ((ch == '"' || ch == 39 || ch == '0' || ch == 92 || ch >= 'a' && ch <= 'b' || ch == 'f' || ch == 'n' || ch == 'r' || ch == 't' || ch == 'v')) {AddCh(); goto case 35;}
				else if (ch == 'x') {AddCh(); goto case 36;}
				else if (ch == 'u') {AddCh(); goto case 38;}
				else if (ch == 'U') {AddCh(); goto case 42;}
				else {t.kind = noSym; break;}
			case 166:
				if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'F' || ch >= 'a' && ch <= 'f')) {AddCh(); goto case 167;}
				else if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '!' || ch >= '#' && ch <= '/' || ch >= ':' && ch <= '@' || ch >= 'G' && ch <= '[' || ch >= ']' && ch <= '`' || ch >= 'g' && ch <= 255)) {AddCh(); goto case 35;}
				else if (ch == '"') {AddCh(); goto case 51;}
				else if (ch == 92) {AddCh(); goto case 165;}
				else {t.kind = noSym; break;}
			case 167:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '!' || ch >= '#' && ch <= '[' || ch >= ']' && ch <= 255)) {AddCh(); goto case 35;}
				else if (ch == '"') {AddCh(); goto case 51;}
				else if (ch == 92) {AddCh(); goto case 165;}
				else {t.kind = noSym; break;}
			case 168:
				if (ch == '"') {AddCh(); goto case 50;}
				else {t.kind = 5; break;}
			case 169:
				if ((ch <= '/' || ch >= ':' && ch <= 255)) {apx++; AddCh(); goto case 2;}
				else if ((ch >= '0' && ch <= '9')) {apx = 0; AddCh(); goto case 10;}
				else {t.kind = noSym; break;}
			case 170:
				if (ch == '=') {AddCh(); goto case 79;}
				else {t.kind = 124; break;}
			case 171:
				if (ch == 'l') {AddCh(); goto case 172;}
				else if (ch == 'n') {AddCh(); goto case 173;}
				else if (ch == 'r') {AddCh(); goto case 110;}
				else {t.kind = noSym; break;}
			case 172:
				if (ch == 'i') {AddCh(); goto case 96;}
				else if (ch == 's') {AddCh(); goto case 99;}
				else {t.kind = noSym; break;}
			case 173:
				if (ch == 'd') {AddCh(); goto case 174;}
				else {t.kind = noSym; break;}
			case 174:
				if (ch == 'i') {AddCh(); goto case 102;}
				else if (ch == 'r') {AddCh(); goto case 130;}
				else {t.kind = noSym; break;}
			case 175:
				{t.kind = 128; break;}
			case 176:
				if (ch == '=') {AddCh(); goto case 72;}
				else if (ch == '-') {AddCh(); goto case 73;}
				else if (ch == '>') {AddCh(); goto case 175;}
				else {t.kind = 111; break;}

		}

		// jetzt kann die Länge des Tokens zu den Code-Zeichen dazuaddiert werden
		if (t.kind != eofSym && t.kind != noSym)
			locTmp.iChCode += tlen;
		
		// Falls sich die Zeile eines Tokens geändert hat, hat eine neue
		// Codezeile begonnen
		if (iOldTokenLine < t.line && t.kind != eofSym && t.kind != noSym) locTmp.iLoCode++;
		iOldTokenLine = t.line;

				
		t.val = new String(tval, 0, tlen);
		return t;
	}
	
	// get the next token (possibly a token already seen during peeking)
	public Token Scan () {
		if (lastTok != null) {
			// Überprüfen, ob der Token ein Operator ist
			if (Parser.ci_operator.Get(lastTok.kind)) {
				if (ciCur.statInfo.htOperators.ContainsKey(Parser.ci_htOperators[lastTok.kind])) {
					int i = (int)ciCur.statInfo.htOperators[Parser.ci_htOperators[lastTok.kind]];
					ciCur.statInfo.htOperators[Parser.ci_htOperators[lastTok.kind]] = i + 1;
				}
				else {
					ciCur.statInfo.htOperators[Parser.ci_htOperators[lastTok.kind]] = 1;
				}		
			}
	
			// Überprüfen, ob der Token ein Schlüsselwort ist
			if (Parser.keyword.Get(lastTok.kind)) {
				if (ciCur.statInfo.htKeywords.ContainsKey(lastTok.val)) {
					int i = (int)ciCur.statInfo.htKeywords[lastTok.val];
					ciCur.statInfo.htKeywords[lastTok.val] = i + 1;
				}
				else {
					ciCur.statInfo.htKeywords[lastTok.val] = 1;
				}
			}

			// Hier wird der Token im Sinne von Halstead klassifiziert
			if (Parser.ci_HalsteadOperands.Get(lastTok.kind)) // gehört zu den Operanden
				ciCur.hInfo.addOperand(lastTok.val);
			else if (!Parser.ci_HalsteadNotCount.Get(lastTok.kind) // gehört nicht zu den Token, die nicht gezählt werden
							&& lastTok.kind != eofSym)
				ciCur	.hInfo.addOperator(lastTok.val);		// gehört zu den Operatoren
	
	
			// Hier werden die gesammelten Daten zum aktuellen Bereich dazuaddiert,
			ciCur.locInfo = ciCur.locInfo + lastTok.loc;
		}
		
		Token ret;
		
		if (tokens.next == null) {
			ret = NextToken();
			ret.loc = locTmp;
			locTmp = new LinesOfCodeInfo();
		} else {
			pt = tokens = tokens.next;
			ret = tokens;
		}
		
		lastTok = ret;
		
		return ret;
	}

	// peek for the next token, ignore pragmas
	public Token Peek () {
		if (pt.next == null) {
			do {
				pt = pt.next = NextToken();
				pt.loc = locTmp;
				locTmp = new LinesOfCodeInfo();
			} while (pt.kind > maxT); // skip pragmas
		} else {
			do {
				pt = pt.next;
			} while (pt.kind > maxT);
		}
		return pt;
	}
	
	// make sure that peeking starts at the current scan position
	public void ResetPeek () { pt = tokens; }

} // end Scanner

}